from typing import Dict
import sqlalchemy as sa
from spinta import commands
from spinta.components import Model, Namespace, Context
from spinta.datasets.components import Dataset
from spinta.exceptions import ModelNotFound, NamespaceNotFound, DatasetNotFound
from spinta.manifests.internal_sql.components import InternalSQLManifest
from spinta.manifests.internal_sql.helpers import internal_to_schema, load_internal_manifest_nodes, get_object_from_id, \
    select_full_table, update_schema_with_external, load_required_models, get_manifest, get_transaction_connection
from spinta.types.namespace import load_namespace_from_name


def get_model_name_list(context: Context, manifest: InternalSQLManifest, loaded: bool, namespace: str = None, recursive: bool = False):
    manifest = get_manifest(context, manifest)
    table = manifest.table
    conn = get_transaction_connection(context)
    objs = manifest.get_objects()
    if conn is None or loaded:
        if 'model' and objs and objs['model']:
            if namespace:
                for model_name, model in objs['model'].items():
                    if model.ns.name == namespace:
                        yield model_name
            else:
                yield from objs['model'].keys()
    else:
        if namespace == '':
            for model in objs['model'].values():
                if model.name.startswith('_'):
                    yield model.name

        if namespace:
            stmt = sa.select(table.c.path).where(
                sa.and_(
                    table.c.path.startswith(namespace),
                    table.c.dim == 'model'
                )
            )
        else:
            stmt = sa.select(table.c.path).where(
                table.c.dim == 'model'
            )
        rows = conn.execute(stmt)
        for row in rows:
            if not recursive and namespace:
                # Check if path is actually right after ns,
                # ex: namespace = 'dataset/test'
                # models: 'dataset/test/gov/Model', 'dataset/test/Model'
                # This will filter out first model, since it belongs to gov namespace
                fixed_path = row['path'].replace(namespace, '')
                if fixed_path.startswith('/'):
                    fixed_path = fixed_path[1:]
                if len(fixed_path.split('/')) == 1:
                    yield row['path']
            else:
                yield row['path']


def get_namespace_name_list(context: Context, manifest: InternalSQLManifest, loaded: bool, namespace: str = None):
    manifest = get_manifest(context, manifest)
    table = manifest.table
    conn = get_transaction_connection(context)
    if conn is None or loaded:
        objs = manifest.get_objects()
        if 'ns' and objs and objs['ns']:
            if namespace is not None:
                for ns_name, ns in objs['ns'].items():
                    if ns.parent and isinstance(ns.parent, Namespace) and ns.parent.name == namespace:
                        yield ns_name
            else:
                yield from objs['ns'].keys()
    else:
        if not namespace:
            stmt = sa.select(table.c.mpath).where(
                sa.or_(
                    table.c.dim == 'namespace',
                    table.c.dim == 'dataset'
                )
            ).order_by(table.c.mpath)
        else:
            stmt = sa.select(table.c.mpath).where(
                sa.and_(
                    table.c.mpath.startswith(namespace),
                    table.c.mpath != namespace,
                    sa.or_(
                        table.c.dim == 'namespace',
                        table.c.dim == 'dataset'
                    )
                )

            ).order_by(table.c.index)
        rows = conn.execute(stmt)
        yielded = []
        for row in rows:
            if namespace:
                # Fix namespace path, ex given namespace is 'dataset/test'
                # Fetched namespaces are: 'dataset/test/gov', 'dataset/test/other/gov'
                # it will return 'dataset/test/gov' and 'dataset/test/other'
                fixed_path = row['mpath'].replace(namespace, '')
                if fixed_path.startswith('/'):
                    fixed_path = fixed_path[1:]
                fixed_path = f'{namespace}/{fixed_path.split("/")[0]}'
                if fixed_path not in yielded:
                    yielded.append(fixed_path)
                    yield fixed_path
            else:
                yield row['mpath']


def _get_dataset_name_list(context: Context, manifest: InternalSQLManifest, loaded: bool):
    manifest = get_manifest(context, manifest)
    table = manifest.table
    conn = get_transaction_connection(context)
    if conn is None or loaded:
        objs = manifest.get_objects()
        if 'dataset' and objs and objs['dataset']:
            yield from objs['dataset'].keys()
    else:
        stmt = sa.select(table.c.path).where(
            table.c.dim == 'dataset'
        ).order_by(table.c.path)
        rows = conn.execute(stmt)
        for row in rows:
            yield row['path']


@commands.has_model.register(Context, InternalSQLManifest, str)
def has_model(context: Context, manifest: InternalSQLManifest, model: str, loaded: bool = False, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    if model in manifest.get_objects()['model']:
        return True
    elif not loaded and conn is not None:
        table = manifest.table
        ns = conn.execute(
            sa.select(table).where(
                sa.and_(
                    table.c.path == model,
                    table.c.dim == 'model'
                )
            )
        )
        if any(ns):
            return True
    return False


@commands.get_model.register(Context, InternalSQLManifest, str)
def get_model(context: Context, manifest: InternalSQLManifest, model: str, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    objects = manifest.get_objects()
    if has_model(context, manifest, model):
        if model in objects['model']:
            m = objects['model'][model]
            return m
        elif conn is not None:
            schemas = []
            table = manifest.table
            m = conn.execute(
                select_full_table(table).where(
                    sa.and_(
                        table.c.path == model,
                        table.c.dim == 'model'
                    )
                ).limit(1)
            )
            model_obj = None
            props = []
            for item in m:
                model_obj = item
                props = conn.execute(
                    select_full_table(table).where(
                        sa.and_(
                            table.c.path.startswith(model),
                            table.c.dim != 'model'
                        )
                    ).order_by(table.c.index)
                )

            parent_id = model_obj['parent']
            parent_dataset = None
            parent_resource = None
            parent_schemas = []
            while parent_id is not None:
                parent_obj = get_object_from_id(context, manifest, parent_id)
                if parent_obj is None:
                    break

                parent_schemas.append(parent_obj)
                parent_id = parent_obj['parent']

                if parent_obj['dim'] == 'dataset':
                    parent_dataset = parent_obj['name']
                    dataset = commands.get_dataset(context, manifest, parent_dataset)
                    if parent_resource:
                        get_dataset_resource(context, manifest, dataset, parent_resource)
                    break
                elif parent_obj['dim'] == 'resource' and not parent_resource:
                    parent_resource = parent_obj['name']

            schemas.extend(reversed(parent_schemas))
            schemas.append(model_obj)
            schemas.extend(props)
            required_models = [model]

            schemas = internal_to_schema(manifest, schemas)
            schemas = update_schema_with_external(schemas, {
                'dataset': parent_dataset,
                'resource': parent_resource
            })
            schemas = load_required_models(context, manifest, schemas, required_models)
            load_internal_manifest_nodes(context, manifest, schemas, link=True, ignore_types=['dataset', 'resource'])
            if model in objects['model']:
                return objects['model'][model]
    raise ModelNotFound(model=model)


@commands.get_models.register(Context, InternalSQLManifest)
def get_models(context: Context, manifest: InternalSQLManifest, loaded: bool = False, **kwargs):
    model_names = get_model_name_list(context, manifest, loaded)
    objs = manifest.get_objects()
    for name in model_names:
        # get_model loads the model if it has not been loaded
        if name not in objs['model']:
            commands.get_model(context, manifest, name)
    return objs['model']


@commands.set_model.register(Context, InternalSQLManifest, str, Model)
def set_model(context: Context, manifest: InternalSQLManifest, model_name: str, model: Model, **kwargs):
    manifest.get_objects()['model'][model_name] = model


@commands.set_models.register(Context, InternalSQLManifest, dict)
def set_models(context: Context, manifest: InternalSQLManifest, models: Dict[str, Model], **kwargs):
    manifest.get_objects()['model'] = models


@commands.has_namespace.register(Context, InternalSQLManifest, str)
def has_namespace(context: Context, manifest: InternalSQLManifest, namespace: str, loaded: bool = False, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    if namespace in manifest.get_objects()['ns']:
        return True
    elif conn is not None and not loaded:
        table = manifest.table
        ns = conn.execute(
            sa.select(table).where(table.c.mpath.startswith(namespace)).limit(1)
        )
        if any(ns):
            return True
    return False


@commands.get_namespace.register(Context, InternalSQLManifest, str)
def get_namespace(context: Context, manifest: InternalSQLManifest, namespace: str, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    objects = manifest.get_objects()

    if has_namespace(context, manifest, namespace):
        if namespace in objects['ns']:
            ns = objects['ns'][namespace]
            return ns
        elif conn is not None:
            table = manifest.table
            ns = conn.execute(
                select_full_table(table).where(
                    sa.and_(
                        table.c.name == namespace,
                        table.c.dim == 'ns'
                    )
                )
            )
            schemas = internal_to_schema(manifest, ns)
            load_internal_manifest_nodes(context, manifest, schemas, link=True)
            if namespace in objects['ns']:
                return objects['ns'][namespace]

            ns = load_namespace_from_name(context, manifest, namespace, drop=False)
            return ns

    raise NamespaceNotFound(namespace=namespace)


@commands.get_namespaces.register(Context, InternalSQLManifest)
def get_namespaces(context: Context, manifest: InternalSQLManifest, loaded: bool = False, **kwargs):
    ns_names = get_namespace_name_list(context, manifest, loaded)
    objs = manifest.get_objects()
    for name in ns_names:
        # get_namespace loads the namespace if it has not been loaded
        if name not in objs['ns']:
            commands.get_namespace(context, manifest, name)
    return objs['ns']


@commands.set_namespace.register(Context, InternalSQLManifest, str, Namespace)
def set_namespace(context: Context, manifest: InternalSQLManifest, namespace: str, ns: Namespace, **kwargs):
    manifest = get_manifest(context, manifest)
    manifest.get_objects()['ns'][namespace] = ns


@commands.has_dataset.register(Context, InternalSQLManifest, str)
def has_dataset(context: Context, manifest: InternalSQLManifest, dataset: str, loaded: bool = False, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    if dataset in manifest.get_objects()['dataset']:
        return True
    elif conn is not None and not loaded:
        table = manifest.table
        ds = conn.execute(
            sa.select(table).where(
                sa.and_(
                    table.c.mpath == dataset,
                    table.c.dim == 'dataset'
                )

            ).limit(1)
        )
        if any(ds):
            return True
    return False


def has_dataset_resource(context: Context, manifest: InternalSQLManifest, dataset: Dataset, resource: str, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    if resource in dataset.resources:
        return True
    elif conn is not None:
        table = manifest.table
        ds = conn.execute(
            sa.select(table).where(
                sa.and_(
                    table.c.path == dataset.name,
                    table.c.dim == 'resource',
                    table.c.name == resource
                )

            ).limit(1)
        )
        if any(ds):
            return True
    return False


def get_dataset_resource(context: Context, manifest: InternalSQLManifest, dataset: Dataset, resource: str, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    if has_dataset_resource(context, manifest, dataset, resource, **kwargs):
        if resource in dataset.resources:
            return dataset.resources[resource]
        elif conn is not None:
            table = manifest.table
            resources = conn.execute(
                sa.select(table).where(
                    sa.and_(
                        table.c.path == dataset.name,
                        sa.or_(
                            sa.and_(
                                table.c.name == resource,
                                table.c.dim == 'resource'
                            ),
                            table.c.dim == 'dataset'
                        )
                    )
                )
            )
            schemas = internal_to_schema(manifest, resources)
            load_internal_manifest_nodes(context, manifest, schemas, link=True)
            if resource in dataset.resources:
                return dataset.resources[resource]


@commands.get_dataset.register(Context, InternalSQLManifest, str)
def get_dataset(context: Context, manifest: InternalSQLManifest, dataset: str, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    objects = manifest.get_objects()

    if has_dataset(context, manifest, dataset):
        if dataset in objects['dataset']:
            return objects['dataset'][dataset]
        elif conn is not None:
            table = manifest.table
            ds = conn.execute(
                select_full_table(table).where(
                    sa.and_(
                        table.c.path == dataset,
                        table.c.dim != 'base',
                        table.c.dim != 'resource',
                    )
                )
            )
            schemas = internal_to_schema(manifest, ds)
            load_internal_manifest_nodes(context, manifest, schemas, link=True)
            if dataset in objects['dataset']:
                return objects['dataset'][dataset]

    raise DatasetNotFound(dataset=dataset)


@commands.get_datasets.register(Context, InternalSQLManifest)
def get_datasets(context: Context, manifest: InternalSQLManifest, loaded: bool = False, **kwargs):
    dataset_names = _get_dataset_name_list(context, manifest, loaded)
    objs = manifest.get_objects()
    for name in dataset_names:
        # get_dataset loads the dataset if it has not been loaded
        if name not in objs['dataset']:
            commands.get_dataset(context, manifest, name)
    return objs['dataset']


@commands.set_dataset.register(Context, InternalSQLManifest, str, Dataset)
def set_dataset(context: Context, manifest: InternalSQLManifest, dataset_name: str, dataset: Dataset, **kwargs):
    manifest.get_objects()['dataset'][dataset_name] = dataset




import asyncio
import json
import os
from typing import List, TypedDict, Dict
from typing import Optional
from typer import Option

import click
from typer import Argument
from typer import Context as TyperContext

from spinta import commands
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.manifest import convert_str_to_manifest_path
from spinta.cli.helpers.store import load_store
from spinta.cli.helpers.store import prepare_manifest
from spinta.components import Context
from spinta.core.context import configure_context
from spinta.exceptions import FileNotFound, ModelNotFound, PropertyNotFound
from spinta.manifests.components import Manifest


def bootstrap(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
):
    """Initialize backends

    This will create tables and sync manifest to backends.
    """
    manifests = convert_str_to_manifest_path(manifests)
    context = configure_context(ctx.obj, manifests)
    store = prepare_manifest(context, ensure_config_dir=True, full_load=True)

    with context:
        require_auth(context)
        commands.bootstrap(context, store.manifest)


def sync(ctx: TyperContext):
    """Sync source manifests into main manifest

    Single main manifest can be populated from multiple different backends.
    """
    context = ctx.obj
    store = load_store(context)
    commands.load(context, store.internal, into=store.manifest)
    with context:
        require_auth(context)
        coro = commands.sync(context, store.manifest)
        asyncio.run(coro)
    click.echo("Done.")


def migrate(
    ctx: TyperContext,
    manifests: Optional[List[str]] = Argument(None, help=(
        "Manifest files to load"
    )),
    plan: bool = Option(False, '-p', '--plan', help=(
        "If added, prints SQL code instead of executing it"
    ), is_flag=True),
    rename: str = Option(None, '-r', '--rename', help=(
        "JSON file, that maps manifest node renaming (models, properties)"
    )),
    autocommit: bool = Option(False, '-a', '--autocommit', help=(
        "If added, migrate will do atomic transactions, meaning it will automatically commit after each action (use it at your own risk)"
    ))
):
    """Migrate schema change to backends"""
    manifests = convert_str_to_manifest_path(manifests)
    context = configure_context(ctx.obj, manifests)
    store = prepare_manifest(context, ensure_config_dir=True)
    with context:
        require_auth(context)
        manifest = store.manifest
        backend = manifest.backend
        migrate_meta = MigrateMeta(
            plan=plan,
            autocommit=autocommit,
            rename=MigrateRename(
                path=rename
            )
        )
        _validate_migrate_rename(context, migrate_meta.rename, manifest)

        if backend:
            context.attach(f'transaction.{backend.name}', backend.begin)
            commands.migrate(context, manifest, backend, migrate_meta)


def freeze(ctx: TyperContext):
    """Detect schema changes and create new version

    This will read current manifest structure, compare it with a previous
    freezed version and will generate new migration version if current and last
    versions do not match.
    """
    context = ctx.obj
    # Load just store, manifests will be loaded by freeze command.
    store = load_store(context)
    with context:
        require_auth(context)
        commands.freeze(context, store.manifest)
    click.echo("Done.")


class MigrateTableRename(TypedDict):
    name: str
    new_name: str
    columns: Dict[str, str]


class MigrateRename:
    tables: Dict[str, MigrateTableRename]

    def __init__(self, path: str):
        self.tables = {}
        self.parse_json_file(path)

    def insert_table(self, table_name: str):
        self.tables[table_name] = MigrateTableRename(
            name=table_name,
            new_name=table_name,
            columns={}
        )

    def insert_column(self, table_name: str, column_name: str, new_column_name: str):
        if table_name not in self.tables.keys():
            self.insert_table(table_name)
        if column_name == "":
            self.tables[table_name]["new_name"] = new_column_name
        else:
            self.tables[table_name]["columns"][column_name] = new_column_name

    def get_column_name(self, table_name: str, column_name: str):
        new_name = column_name.split(".")
        if table_name in self.tables.keys():
            table = self.tables[table_name]
            if new_name[0] in table["columns"].keys():
                new_name[0] = table["columns"][new_name[0]]
                return '.'.join(new_name)
        return column_name

    def get_old_column_name(self, table_name: str, column_name: str):
        new_name = column_name.split(".")
        if table_name in self.tables.keys():
            table = self.tables[table_name]
            for column, new_column_name in table["columns"].items():
                if new_column_name == new_name[0]:
                    new_name[0] = column
                    return '.'.join(new_name)
        return column_name

    def get_table_name(self, table_name: str):
        if table_name in self.tables.keys():
            return self.tables[table_name]["new_name"]
        return table_name

    def get_old_table_name(self, table_name: str):
        for key, data in self.tables.items():
            if data["new_name"] == table_name:
                return key
        return table_name

    def parse_json_file(self, path: str):
        if path:
            if os.path.exists(path):
                with open(path, 'r') as f:
                    data = json.loads(f.read())
                    for table, table_data in data.items():
                        self.insert_table(table)
                        for column, column_data in table_data.items():
                            self.insert_column(table, column, column_data)
            else:
                raise FileNotFound(file=path)


class MigrateMeta:
    plan: bool
    autocommit: bool
    rename: MigrateRename

    def __init__(self, plan: bool, autocommit: bool, rename: MigrateRename):
        self.plan = plan
        self.rename = rename
        self.autocommit = autocommit


def _validate_migrate_rename(context: Context, rename: MigrateRename, manifest: Manifest):
    tables = rename.tables.values()
    for table in tables:
        models = commands.get_models(context, manifest)
        if table["new_name"] not in models.keys():
            raise ModelNotFound(model=table["new_name"])
        model = commands.get_model(context, manifest, table["new_name"])
        for column in table["columns"].values():
            if column not in model.properties.keys():
                raise PropertyNotFound(property=column)



import uuid
from copy import deepcopy
from dataclasses import dataclass
from typing import overload, Optional, Iterator, List, Tuple, Callable, TypedDict
from pathlib import Path

from starlette.requests import Request
from starlette.responses import Response
from starlette.responses import FileResponse

from spinta import commands
from spinta.backends.helpers import get_select_prop_names
from spinta.backends.helpers import get_select_tree
from spinta.backends.components import Backend
from spinta.backends.nobackend.components import NoBackend
from spinta.compat import urlparams_to_expr
from spinta.components import Context, Node, Action, UrlParams, Page, PageBy, get_page_size
from spinta.components import Model
from spinta.components import Property
from spinta.core.ufuncs import Expr
from spinta.datasets.components import ExternalBackend
from spinta.renderer import render
from spinta.types.datatype import Integer
from spinta.types.datatype import Object
from spinta.types.datatype import File
from spinta.accesslog import AccessLog
from spinta.accesslog import log_response
from spinta.exceptions import UnavailableSubresource, InfiniteLoopWithPagination, BackendNotGiven, TooShortPageSize, \
    TooShortPageSizeKeyRepetition
from spinta.exceptions import ItemDoesNotExist
from spinta.types.datatype import DataType
from spinta.typing import ObjectData
from spinta.ufuncs.basequerybuilder.components import get_allowed_page_property_types, QueryParams, \
    update_query_with_url_params
from spinta.ufuncs.basequerybuilder.ufuncs import add_page_expr
from spinta.utils.data import take


@commands.getall.register(Context, Model, Request)
async def getall(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)
    backend = model.backend
    if isinstance(backend, NoBackend):
        raise BackendNotGiven(model)

    copy_page = prepare_page_for_get_all(context, model, params)
    is_page_enabled = not params.count and backend.paginated and copy_page and copy_page and copy_page.is_enabled

    if isinstance(backend, ExternalBackend):
        # XXX: `add_count` is a hack, because, external backends do not
        #      support it yet.
        expr = urlparams_to_expr(params, add_count=False)
    else:
        expr = urlparams_to_expr(params)
    query_params = QueryParams()
    update_query_with_url_params(query_params, params)
    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
    )

    if params.head:
        rows = []
    else:
        if is_page_enabled:
            rows = get_page(context, model, backend, copy_page, expr, params.limit, default_expand=False, params=query_params)
        else:
            if backend.support_expand:
                rows = commands.getall(context, model, backend, params=query_params, query=expr, default_expand=False)
            else:
                rows = commands.getall(context, model, backend, params=query_params, query=expr)

    if params.count:
        # XXX: Quick and dirty hack. Functions should be handled properly.
        prop = Property()
        prop.name = 'count()'
        prop.place = 'count()'
        prop.title = ''
        prop.description = ''
        prop.model = model
        prop.dtype = Integer()
        prop.dtype.type = 'integer'
        prop.dtype.type_args = []
        prop.dtype.name = 'integer'
        prop.dtype.prop = prop
        props = {
            '_type': model.properties['_type'],
            'count()': prop,
        }
        rows = (
            {
                prop.name: commands.prepare_dtype_for_response(
                    context,
                    params.fmt,
                    props[key].dtype,
                    val,
                    data=row,
                    action=action,
                )
                for key, val in row.items() if key in props
            }
            for row in rows
        )
    else:
        select_tree = get_select_tree(context, action, params.select)
        if action == Action.SEARCH:
            reserved = ['_type', '_id', '_revision', '_base']
        else:
            reserved = ['_type', '_id', '_revision']
        if model.page.is_enabled:
            reserved.append('_page')
        prop_names = get_select_prop_names(
            context,
            model,
            model.properties,
            action,
            select_tree,
            reserved=reserved,
            include_denorm_props=False,
        )
        rows = (
            commands.prepare_data_for_response(
                context,
                model,
                params.fmt,
                row,
                action=action,
                select=select_tree,
                prop_names=prop_names,
            )
            for row in rows
        )

    rows = log_response(context, rows)

    return render(context, request, model, params, rows, action=action)


# XXX: params.sort handle should be moved to BaseQueryBuilder, AST should be handled by Env, this only supports basic sort arguments, that ar Positive, Negative or Bind.
# Issue: in order to create correct WHERE, we need to have finalized Page, currently we combine sort properties with page here.
def prepare_page_for_get_all(context: Context, model: Model, params: UrlParams):
    if model.page:
        copied = deepcopy(model.page)
        copied.clear()

        if params.page:
            copied.is_enabled = params.page.is_enabled
        if copied.is_enabled:
            config = context.get('config')
            page_size = config.push_page_size
            size = params.page and params.page.size or copied.size or page_size or 1000
            copied.size = size

            if params.sort:
                new_order = {}
                for sort_ast in params.sort:
                    negative = False
                    if sort_ast['name'] == 'negative':
                        negative = True
                        sort_ast = sort_ast['args'][0]
                    elif sort_ast['name'] == 'positive':
                        sort_ast = sort_ast['args'][0]
                    for sort in sort_ast['args']:
                        if isinstance(sort, str):
                            if sort in model.properties.keys():
                                by = sort if not negative else f'-{sort}'
                                prop = model.properties[sort]
                                if isinstance(prop.dtype, get_allowed_page_property_types()):
                                    new_order[by] = PageBy(prop)
                                else:
                                    copied.is_enabled = False
                                    break
                        else:
                            copied.is_enabled = False
                            break

                for key, page_by in copied.by.items():
                    reversed_key = key[1:] if key.startswith("-") else f'-{key}'
                    if key not in new_order and reversed_key not in new_order:
                        new_order[key] = page_by

                copied.by = new_order

            if params.page and params.page.values:
                copied.update_values_from_list(params.page.values)

        return copied


def _is_iter_last_real_value(it: int, total: int, added_size: int = 1):
    return it > (total - 1 - added_size)


def _is_iter_last_potential_value(it: int, total: int):
    return it > (total - 1)


def get_page(
    context: Context,
    model: Model,
    backend: Backend,
    model_page: Page,
    expr: Expr,
    limit: Optional[int] = None,
    default_expand: bool = True,
    params: QueryParams = None,
) -> Iterator[ObjectData]:
    config = context.get('config')
    size = get_page_size(config, model, model_page)

    # Add 1 to see future value (to see if it finished, check for infinite loops and page size miss matches).
    model_page.size = size + 1

    page_meta = PaginationMetaData(
        page_size=size,
        limit=limit
    )
    while not page_meta.is_finished:
        page_meta.is_finished = True
        query = add_page_expr(expr, model_page)
        if backend.support_expand:
            rows = commands.getall(context, model, backend, params=params, query=query, default_expand=default_expand)
        else:
            rows = commands.getall(context, model, backend, params=params, query=query)

        yield from get_paginated_values(model_page, page_meta, rows)


@dataclass
class PaginationMetaData:
    page_size: int
    previous_first_value = None
    is_first_iter = True
    is_finished = False
    true_count = 0
    limit: int = None

    def handle_count(self) -> bool:
        if self.limit:
            if self.true_count >= self.limit:
                self.is_finished = True
                return True
            self.true_count += 1
        return False


def get_paginated_values(model_page: Page, meta: PaginationMetaData, rows):
    size = meta.page_size

    if model_page.first_time != meta.is_first_iter:
        model_page.first_time = meta.is_first_iter
    if meta.is_first_iter:
        meta.is_first_iter = False

    initial_key = [val.value for val in model_page.by.values()]
    current_first_value = None
    previous_value = None
    key_repetition = [[], 0]
    flag_for_potential_key_repetition = False
    for i, row in enumerate(rows):
        # Check if future value is not the same as last value
        if '_page' in row:
            previous_key = row['_page'].copy()
            if key_repetition[0] == previous_key:
                key_repetition[1] += 1
            else:
                key_repetition = [previous_key, 0]

        if _is_iter_last_potential_value(i, size):
            meta.is_finished = False
            if flag_for_potential_key_repetition:
                raise TooShortPageSize(
                    model_page,
                    page_size=size,
                    page_values=previous_value
                )
            if key_repetition[1] > 0:
                raise TooShortPageSizeKeyRepetition(
                    model_page,
                    page_size=size,
                    page_values=previous_value,
                )
            break

        previous_value = row

        limit_result = meta.handle_count()
        if limit_result:
            break

        # Check if row is completely the same as previous page first row
        if current_first_value is None:
            current_first_value = row
            if current_first_value == meta.previous_first_value:
                raise InfiniteLoopWithPagination(
                    model_page,
                    page_size=size,
                    page_values=current_first_value
                )
            meta.previous_first_value = current_first_value

        if '_page' in row:
            model_page.update_values_from_list(row['_page'])

            # Check if initial key is the same as last key
            if _is_iter_last_real_value(i, size):
                last_key = row['_page'].copy()
                if initial_key == last_key:
                    flag_for_potential_key_repetition = True

        yield row


def _update_expr_args(
    expr: Expr,
    name: str,
    args: List,
    override: bool = False
) -> Tuple[bool, Expr]:
    updated = False
    expr.args = list(expr.args)
    if expr.name == name:
        if override:
            expr.args = args
            updated = True
        else:
            if expr.args:
                expr.args.extend(args)
                updated = True
    else:
        for i, arg in enumerate(expr.args):
            if isinstance(arg, Expr):
                updated, expr.args[i] = _update_expr_args(arg, name, args, override)
                if updated:
                    break
    expr.args = tuple(expr.args)
    return updated, expr


@overload
@commands.getone.register(Context, Request, Node)
async def getone(
    context: Context,
    request: Request,
    node: Node,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    if params.prop and params.propref:
        resp = await commands.getone(
            context,
            request,
            params.prop,
            params.prop.dtype,
            params.model.backend,
            action=action,
            params=params,
        )
    elif params.prop:
        resp = await commands.getone(
            context,
            request,
            params.prop,
            params.prop.dtype,
            params.prop.dtype.backend or params.model.backend,
            action=action,
            params=params,
        )
    else:
        resp = await commands.getone(
            context,
            request,
            params.model,
            params.model.backend,
            action=action,
            params=params,
        )

    accesslog: AccessLog = context.get('accesslog')
    accesslog.response(objects=1)

    return resp


@overload
@commands.getone.register(Context, Request, Property, DataType, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: DataType,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    raise UnavailableSubresource(prop=prop.name, prop_type=prop.dtype.name)


@overload
@commands.getone.register(Context, Request, Model, Backend)
async def getone(
    context: Context,
    request: Request,
    model: Model,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
        id_=params.pk,
    )

    data = commands.getone(context, model, backend, id_=params.pk)
    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
    )
    data = commands.prepare_data_for_response(
        context,
        model,
        params.fmt,
        data,
        action=action,
        select=select_tree,
        prop_names=prop_names,
    )
    return render(context, request, model, params, data, action=action)


@overload
@commands.getone.register(Context, Request, Property, Object, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: Object,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    data = commands.getone(
        context,
        prop,
        dtype,
        backend,
        id_=params.pk,
    )

    data = commands.prepare_data_for_response(
        context,
        prop.dtype,
        params.fmt,
        data,
        action=action,
    )
    return render(context, request, prop, params, data, action=action)


@overload
@commands.getone.register(Context, Request, Property, File, Backend)
async def getone(
    context: Context,
    request: Request,
    prop: Property,
    dtype: File,
    backend: Backend,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, prop)

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=prop.model.model_type(),
        prop=prop.place,
        action=action.value,
        id_=params.pk,
    )

    # Get metadata from model backend
    data = commands.getone(
        context,
        prop,
        dtype,
        prop.model.backend,
        id_=params.pk,
    )

    # Return file metadata from model backend
    if params.propref:
        data = commands.prepare_data_for_response(
            context,
            prop.dtype,
            params.fmt,
            data,
            action=Action.GETONE,
        )
        return render(context, request, prop, params, data, action=action)

    # Return file content from property backend
    else:
        value = take(prop.place, data)

        file = commands.getfile(
            context,
            prop,
            dtype,
            dtype.backend,
            data=value,
        )

        if file is None:
            raise ItemDoesNotExist(dtype, id=params.pk)

        filename = value['_id']

        if isinstance(file, bytes):
            ResponseClass = Response
        elif isinstance(file, Path):
            ResponseClass = FileResponse
        else:
            raise ValueError(f"Unknown file type {type(file)}.")

        return ResponseClass(
            file,
            media_type=value.get('_content_type'),
            headers={
                'Revision': data['_revision'],
                'Content-Disposition': (
                    f'attachment; filename="{filename}"'
                    if filename else
                    'attachment'
                )
            },
        )


@commands.changes.register(Context, Model, Request)
async def changes(
    context: Context,
    model: Model,
    request: Request,
    *,
    action: Action,
    params: UrlParams,
):
    commands.authorize(context, action, model)

    if params.head:
        rows = []
    else:
        accesslog = context.get('accesslog')
        accesslog.request(
            # XXX: Read operations does not have a transaction, but it
            #      is needed for loging.
            txn=str(uuid.uuid4()),
            model=model.model_type(),
            action=action.value,
        )

        rows = commands.changes(
            context,
            model,
            model.backend,
            id_=params.pk,
            limit=params.limit,
            offset=params.changes_offset,
        )

    select_tree = get_select_tree(context, action, params.select)
    prop_names = get_select_prop_names(
        context,
        model,
        model.properties,
        action,
        select_tree,
        reserved=[
            '_cid',
            '_created',
            '_op',
            '_id',
            '_txn',
            '_revision',
        ],
    )
    rows = (
        commands.prepare_data_for_response(
            context,
            model,
            params.fmt,
            row,
            action=action,
            select=select_tree,
            prop_names=prop_names,
        )
        for row in rows
    )

    return render(context, request, model, params, rows, action=action)


@commands.summary.register(Context, Request, Model)
async def summary(
    context: Context,
    request: Request,
    model: Model,
    *,
    action: Action,
    params: UrlParams,
) -> Response:
    commands.authorize(context, action, model)
    backend = model.backend

    accesslog: AccessLog = context.get('accesslog')
    accesslog.request(
        # XXX: Read operations does not have a transaction, but it
        #      is needed for loging.
        txn=str(uuid.uuid4()),
        model=model.model_type(),
        action=action.value,
    )
    if params.head:
        rows = []
    else:
        expr = urlparams_to_expr(params)
        rows = commands.summary(context, model, backend, expr)
    rows = log_response(context, rows)
    return render(context, request, model, params, rows, action=action)



from typing import List

from spinta import commands
from spinta.backends.constants import TableType
from spinta.backends.postgresql.components import PostgreSQL
from spinta.backends.postgresql.helpers import get_column_name as gcn
from spinta.components import Property, Model
from spinta.datasets.enums import Level
from spinta.exceptions import PropertyNotFound
from spinta.types.datatype import DataType, Ref, BackRef, String, ExternalRef
import sqlalchemy as sa

from spinta.types.text.components import Text


def convert_str_to_columns(table: sa.Table, prop: Property, columns: List[str]) -> List[sa.Column]:
    return_list = []
    for column in columns:
        return_list.append(convert_str_to_column(table, prop, column))
    return return_list


def convert_str_to_column(table: sa.Table, prop: Property, column: str) -> sa.Column:
    if column in table.c:
        return table.c[column]
    else:
        raise PropertyNotFound(prop.dtype)


def get_table(backend: PostgreSQL, prop: Property) -> sa.Table:
    if prop.list:
        table = backend.get_table(prop.list, TableType.LIST)
    else:
        table = backend.get_table(prop)
    return table


@commands.get_column.register(PostgreSQL, Property)
def get_column(backend: PostgreSQL, prop: Property, **kwargs):
    return commands.get_column(backend, prop.dtype, **kwargs)


@commands.get_column.register(PostgreSQL, DataType)
def get_column(backend: PostgreSQL, dtype: DataType, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)
    column_name = gcn(dtype.prop)
    return convert_str_to_column(table, prop, column_name)


@commands.get_column.register(PostgreSQL, String)
def get_column(backend: PostgreSQL, dtype: String, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    parent = prop.parent

    if table is None:
        table = get_table(backend, prop)
    column_name = gcn(parent if parent and isinstance(parent.dtype, Text) else dtype.prop)
    column = convert_str_to_column(table, prop, column_name)

    if parent and isinstance(parent.dtype, Text):
        column = column[dtype.prop.name].label(dtype.prop.place.replace('@', '.'))
    return column


@commands.get_column.register(PostgreSQL, Text)
def get_column(backend: PostgreSQL, dtype: Text, table: sa.Table = None, langs: list = [], default_langs: list = [],
               push: bool = False, **kwargs):
    prop = dtype.prop
    column_name = gcn(dtype.prop)
    if table is None:
        table = get_table(backend, prop)

    if push:
        column = convert_str_to_column(table, prop, column_name)
        return column

    existing_langs = list(dtype.langs.keys())
    lang_prop = None
    if langs:
        for lang in langs:
            if lang in dtype.langs:
                lang_prop = dtype.langs[lang]
                break
    if not lang_prop and default_langs:
        for lang in default_langs:
            if lang in dtype.langs:
                lang_prop = dtype.langs[lang]
                break
    if not lang_prop:
        lang_prop = dtype.langs[existing_langs[0]]
    column = convert_str_to_column(table, prop, column_name)
    column = column[lang_prop.name].label(prop.place)
    return column


@commands.get_column.register(PostgreSQL, Ref)
def get_column(backend: PostgreSQL, dtype: Ref, table: sa.Table = None, **kwargs):
    column = gcn(dtype.prop, replace=not dtype.prop.given.explicit)
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)
    if prop.list is not None:
        column = '_id'
    else:
        column += '._id'
    return convert_str_to_column(table, prop, column)


@commands.get_column.register(PostgreSQL, ExternalRef)
def get_column(backend: PostgreSQL, dtype: Ref, table: sa.Table = None, **kwargs):
    column_prefix = gcn(dtype.prop, replace=not dtype.prop.given.explicit)
    prop = dtype.prop
    if table is None:
        table = get_table(backend, prop)

    result = []
    for ref_prop in dtype.refprops:
        ref_prop_ = ref_prop
        table_ = table
        column = ref_prop.name
        if prop.list is None:
            column = f'{column_prefix}.{ref_prop.name}'
        if ref_prop.name in dtype.properties:
            if isinstance(ref_prop.dtype, type(dtype.properties[ref_prop.name].dtype)):
                ref_prop_ = dtype.properties[ref_prop.name]
                table_ = get_table(backend, ref_prop_)
                column = ref_prop_.place
        result.append(convert_str_to_column(table_, ref_prop_, column))
    return result


@commands.get_column.register(PostgreSQL, BackRef)
def get_column(backend: PostgreSQL, dtype: BackRef, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    r_prop = dtype.refprop
    columns = []
    if table is None:
        table = get_table(backend, dtype.prop)
    if r_prop.level is None or r_prop.level > Level.open:
        column = '_id'
        return convert_str_to_column(table, prop, column)
    else:
        for refprop in r_prop.dtype.refprops:
            columns.append(refprop.name)
        return convert_str_to_columns(table, prop, columns)


@commands.get_column.register(PostgreSQL, Property, Model)
def get_column(backend: PostgreSQL, prop: Property, model: Model, table: sa.Table = None, **kwargs):
    return commands.get_column(backend, prop.dtype, model, table=table, **kwargs)


@commands.get_column.register(PostgreSQL, DataType, Model)
def get_column(backend: PostgreSQL, prop: Property, model: Model, table: sa.Table = None, **kwargs):
    return None


@commands.get_column.register(PostgreSQL, Ref, Model)
def get_column(backend: PostgreSQL, dtype: Ref, model: Model, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = backend.get_table(model)
    column = '_id'
    return convert_str_to_column(table, prop, column)


@commands.get_column.register(PostgreSQL, ExternalRef, Model)
def get_column(backend: PostgreSQL, dtype: ExternalRef, model: Model, table: sa.Table = None, **kwargs):
    prop = dtype.prop
    if table is None:
        table = backend.get_table(model)

    columns = []
    for refprop in prop.dtype.refprops:
        columns.append(refprop.name)
    return convert_str_to_columns(table, prop, columns)






from abc import ABC, abstractmethod
from typing import List

import sqlalchemy as sa

from alembic.operations import Operations


class MigrationAction(ABC):
    @abstractmethod
    def execute(self, op: Operations):
        pass


class CreateTableMigrationAction(MigrationAction):
    def __init__(self, table_name: str, columns: List[sa.Column]):
        self.table_name = table_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_table(self.table_name, *self.columns)


class DropTableMigrationAction(MigrationAction):
    def __init__(self, table_name: str):
        self.table_name = table_name

    def execute(self, op: Operations):
        op.drop_table(table_name=self.table_name)


class RenameTableMigrationAction(MigrationAction):
    def __init__(self, old_table_name: str, new_table_name: str):
        self.old_table_name = old_table_name
        self.new_table_name = new_table_name

    def execute(self, op: Operations):
        op.rename_table(old_table_name=self.old_table_name, new_table_name=self.new_table_name)


class AddColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column: sa.Column):
        self.table_name = table_name
        self.column = column

    def execute(self, op: Operations):
        op.add_column(table_name=self.table_name, column=self.column)


class DropColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column_name: str):
        self.table_name = table_name
        self.column_name = column_name

    def execute(self, op: Operations):
        op.drop_column(table_name=self.table_name, column_name=self.column_name)


class AlterColumnMigrationAction(MigrationAction):
    def __init__(self, table_name: str, column_name: str, nullable: bool = None, new_column_name: str = None,
                 type_=None, using: str = None):
        self.table_name = table_name
        self.column_name = column_name
        self.nullable = nullable
        self.new_column_name = new_column_name
        self.type_ = type_
        self.using = using

    def execute(self, op: Operations):
        op.alter_column(
            table_name=self.table_name,
            column_name=self.column_name,
            nullable=self.nullable,
            new_column_name=self.new_column_name,
            type_=self.type_,
            postgresql_using=self.using
        )


class DropConstraintMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str):
        self.table_name = table_name
        self.constraint_name = constraint_name

    def execute(self, op: Operations):
        op.drop_constraint(constraint_name=self.constraint_name, table_name=self.table_name)


class DropIndexMigrationAction(MigrationAction):
    def __init__(self, table_name: str, index_name: str):
        self.table_name = table_name
        self.index_name = index_name

    def execute(self, op: Operations):
        op.drop_index(index_name=self.index_name, table_name=self.table_name)


class CreateUniqueConstraintMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str, columns: List[str]):
        self.table_name = table_name
        self.constraint_name = constraint_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_unique_constraint(constraint_name=self.constraint_name, table_name=self.table_name, columns=self.columns)


class CreatePrimaryKeyMigrationAction(MigrationAction):
    def __init__(self, table_name: str, constraint_name: str, columns: List[str]):
        self.table_name = table_name
        self.constraint_name = constraint_name
        self.columns = columns

    def execute(self, op: Operations):
        op.create_primary_key(constraint_name=self.constraint_name, table_name=self.table_name, columns=self.columns)


class CreateIndexMigrationAction(MigrationAction):
    def __init__(self, table_name: str, index_name: str, columns: List[str], using: str = None):
        self.table_name = table_name
        self.index_name = index_name
        self.columns = columns
        self.using = using

    def execute(self, op: Operations):
        op.create_index(index_name=self.index_name, table_name=self.table_name, columns=self.columns, postgresql_using=self.using)


class DowngradeTransferDataMigrationAction(MigrationAction):
    def __init__(self, table_name: str, foreign_table_name: str, columns: List[str]):
        set_string = []
        main_ref = ""
        for column in columns:
            split = column.split(".")
            main_ref = '.'.join(split[:-1])
            split = split[-1]
            set_string.append(f'\n"{main_ref}.{split}" = new."{split}"')

        self.query = f'''
        UPDATE "{table_name}" AS old
        SET {','.join(set_string)}
        FROM "{foreign_table_name}" as new
        WHERE old."{main_ref}._id" = new."_id"
        '''

    def execute(self, op: Operations):
        op.execute(self.query)


class UpgradeTransferDataMigrationAction(MigrationAction):
    def __init__(self, table_name: str, foreign_table_name: str, columns: List[str]):
        where_string = []
        main_ref = ""
        for column in columns:
            split = column.split(".")
            main_ref = '.'.join(split[:-1])
            split = split[-1]
            where_string.append(f'\nold."{main_ref}.{split}" = new."{split}"')

        self.query = f'''
        UPDATE "{table_name}" AS old
        SET "{main_ref}._id" = new."_id"
        FROM "{foreign_table_name}" AS new
        WHERE {' AND '.join(where_string)}
        '''

    def execute(self, op: Operations):
        op.execute(self.query)


class CreateForeignKeyMigrationAction(MigrationAction):
    def __init__(self, source_table: str, referent_table: str, constraint_name: str, local_cols: list, remote_cols: list):
        self.constraint_name = constraint_name
        self.source_table = source_table
        self.referent_table = referent_table
        self.local_cols = local_cols
        self.remote_cols = remote_cols

    def execute(self, op: Operations):
        op.create_foreign_key(
            constraint_name=self.constraint_name,
            source_table=self.source_table,
            referent_table=self.referent_table,
            local_cols=self.local_cols,
            remote_cols=self.remote_cols
        )


class RenameSequenceMigrationAction(MigrationAction):
    def __init__(self, old_name: str, new_name: str):
        self.query = f'ALTER SEQUENCE "{old_name}" RENAME TO "{new_name}"'

    def execute(self, op: Operations):
        replaced = self.query.replace(":", "\:")
        op.execute(replaced)


class MigrationHandler:
    def __init__(self):
        self.migrations: List[MigrationAction] = []
        self.foreign_key_migration: List[MigrationAction] = []

    def add_action(self, action: MigrationAction, foreign_key: bool = False):
        if foreign_key:
            self.foreign_key_migration.append(action)
        else:
            self.migrations.append(action)

    def has_constraint_been_dropped(self, constraint_name: str):
        for migration in self.migrations:
            if isinstance(migration, DropConstraintMigrationAction):
                if migration.constraint_name == constraint_name:
                    return True
        for migration in self.foreign_key_migration:
            if isinstance(migration, DropConstraintMigrationAction):
                if migration.constraint_name == constraint_name:
                    return True
        return False

    def run_migrations(self, op: Operations):
        for migration in self.migrations:
            migration.execute(op)
        for migration in self.foreign_key_migration:
            migration.execute(op)



from typing import TypedDict, Callable, Dict

from spinta import commands
from spinta.components import Namespace, Model, Node, Context
from spinta.datasets.components import Dataset
from spinta.exceptions import DatasetNotFound, NamespaceNotFound, ModelNotFound, ManifestObjectNotDefined
from spinta.manifests.components import Manifest


class _FunctionTypes(TypedDict):
    has: Callable
    get: Callable
    set: Callable
    get_all: Callable


NODE_FUNCTION_MAPPER = {
    'model': _FunctionTypes(
        has=commands.has_model,
        get=commands.get_model,
        set=commands.set_model,
        get_all=commands.get_models
    ),
    'ns': _FunctionTypes(
        has=commands.has_namespace,
        get=commands.get_namespace,
        set=commands.set_namespace,
        get_all=commands.get_namespaces
    ),
    'dataset': _FunctionTypes(
        has=commands.has_dataset,
        get=commands.get_dataset,
        set=commands.set_dataset,
        get_all=commands.get_datasets
    )
}


@commands.has_node_type.register(Context, Manifest, str)
def has_object_type(context: Context, manifest: Manifest, obj_type: str, **kwargs):
    return obj_type in manifest.get_objects()


@commands.has_node.register(Context, Manifest, str, str)
def has_object(context: Context, manifest: Manifest, obj_type: str, obj: str, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['has'](context, manifest, obj, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.get_node.register(Context, Manifest, str, str)
def get_node(context: Context, manifest: Manifest, obj_type: str, obj: str, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['get'](context, manifest, obj, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.get_nodes.register(Context, Manifest, str)
def get_nodes(context: Context, manifest: Manifest, obj_type: str, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['get_all'](context, manifest, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.set_node.register(Context, Manifest, str, str, Node)
def set_node(context: Context, manifest: Manifest, obj_type: str, obj_name, obj: Node, **kwargs):
    if obj_type in NODE_FUNCTION_MAPPER:
        return NODE_FUNCTION_MAPPER[obj_type]['set'](context, manifest, obj_name, obj, **kwargs)
    raise ManifestObjectNotDefined(obj=obj_type)


@commands.has_model.register(Context, Manifest, str)
def has_model(context: Context, manifest: Manifest, model: str, **kwargs):
    return model in manifest.get_objects()['model']


@commands.get_model.register(Context, Manifest, str)
def get_model(context: Context, manifest: Manifest, model: str, **kwargs):
    if has_model(context, manifest, model):
        return manifest.get_objects()['model'][model]
    raise ModelNotFound(model=model)


@commands.get_models.register(Context, Manifest)
def get_models(context: Context, manifest: Manifest, **kwargs):
    return manifest.get_objects()['model']


@commands.set_model.register(Context, Manifest, str, Model)
def set_model(context: Context, manifest: Manifest, model_name: str, model: Model, **kwargs):
    manifest.get_objects()['model'][model_name] = model


@commands.set_models.register(Context, Manifest, dict)
def set_models(context: Context, manifest: Manifest, models: Dict[str, Model], **kwargs):
    manifest.get_objects()['model'] = models


@commands.has_namespace.register(Context, Manifest, str)
def has_namespace(context: Context, manifest: Manifest, namespace: str, **kwargs):
    return namespace in manifest.get_objects()['ns']


@commands.get_namespace.register(Context, Manifest, str)
def get_namespace(context: Context, manifest: Manifest, namespace: str, **kwargs):
    if has_namespace(context, manifest, namespace):
        return manifest.get_objects()['ns'][namespace]
    raise NamespaceNotFound(namespace=namespace)


@commands.get_namespaces.register(Context, Manifest)
def get_namespaces(context: Context, manifest: Manifest, **kwargs):
    return manifest.get_objects()['ns']


@commands.set_namespace.register(Context, Manifest, str, Namespace)
def set_namespace(context: Context, manifest: Manifest, namespace: str, ns: Namespace, **kwargs):
    manifest.get_objects()['ns'][namespace] = ns


@commands.has_dataset.register(Context, Manifest, str)
def has_dataset(context: Context, manifest: Manifest, dataset: str, **kwargs):
    return dataset in manifest.get_objects()['dataset']


@commands.get_dataset.register(Context, Manifest, str)
def get_dataset(context: Context, manifest: Manifest, dataset: str, **kwargs):
    if has_dataset(context, manifest, dataset):
        return manifest.get_objects()['dataset'][dataset]
    raise DatasetNotFound(dataset=dataset)


@commands.get_datasets.register(Context, Manifest)
def get_datasets(context: Context, manifest: Manifest, **kwargs):
    return manifest.get_objects()['dataset']


@commands.set_dataset.register(Context, Manifest, str, Dataset)
def set_dataset(context: Context, manifest: Manifest, dataset_name: str, dataset: Dataset, **kwargs):
    manifest.get_objects()['dataset'][dataset_name] = dataset



from typing import Any, List

import geoalchemy2.types

from spinta.backends.components import BackendFeatures
from spinta.backends.constants import TableType
from spinta.backends.helpers import get_table_name
from spinta.backends.postgresql.helpers import get_pg_name, get_column_name
from spinta.backends.postgresql.helpers.migrate import MigrationHandler
from spinta.datasets.inspect.helpers import zipitems
import spinta.backends.postgresql.helpers.migrate as ma

import sqlalchemy as sa
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.dialects.postgresql import JSONB, BIGINT, ARRAY
from sqlalchemy.dialects import postgresql

from spinta.cli.migrate import MigrateMeta, MigrateRename
from spinta.commands import create_exception
from spinta.datasets.enums import Level
from spinta.types.datatype import DataType, Ref, File, Inherit, Array, Object
from spinta.types.namespace import sort_models_by_ref_and_base
from spinta.utils.schema import NotAvailable, NA

from alembic.migration import MigrationContext
from alembic.operations import Operations

from spinta import commands
from spinta.components import Context, Model, Property
from spinta.manifests.components import Manifest
from spinta.backends.postgresql.components import PostgreSQL

EXCLUDED_MODELS = (
    'spatial_ref_sys'
)


@commands.migrate.register(Context, Manifest, PostgreSQL, MigrateMeta)
def migrate(context: Context, manifest: Manifest, backend: PostgreSQL, migrate_meta: MigrateMeta):
    conn = context.get(f'transaction.{backend.name}')
    ctx = MigrationContext.configure(conn, opts={
        "as_sql": migrate_meta.plan
    })
    op = Operations(ctx)
    inspector = sa.inspect(conn)
    table_names = inspector.get_table_names()
    metadata = sa.MetaData(bind=conn)
    metadata.reflect()

    tables = []
    models = commands.get_models(context, manifest)
    for table in table_names:
        name = migrate_meta.rename.get_table_name(table)
        if name not in models.keys():
            name = table
        tables.append(name)
    sorted_models = sort_models_by_ref_and_base(list(models.values()))
    sorted_model_names = list([model.name for model in sorted_models])
    # Do reversed zip, to ensure that sorted models get selected first
    models = zipitems(
        sorted_model_names,
        tables,
        _model_name_key
    )
    handler = MigrationHandler()
    for items in models:
        for new_model, old_model in items:
            if old_model and any(value in old_model for value in (TableType.CHANGELOG.value, TableType.FILE.value)):
                continue
            if old_model and old_model in EXCLUDED_MODELS:
                continue
            old = NA
            if old_model:
                old = metadata.tables[migrate_meta.rename.get_old_table_name(old_model)]
            new = commands.get_model(context, manifest, new_model) if new_model else new_model
            commands.migrate(context, backend, inspector, old, new, handler, migrate_meta.rename)
    _handle_foreign_key_constraints(inspector, sorted_models, handler, migrate_meta.rename)
    _clean_up_file_type(inspector, sorted_models, handler, migrate_meta.rename)
    try:
        if migrate_meta.autocommit:
            # Recreate new connection, that auto commits
            with backend.engine.connect().execution_options(isolation_level="AUTOCOMMIT") as conn:
                ctx = MigrationContext.configure(conn, opts={
                    "as_sql": migrate_meta.plan
                })
                op = Operations(ctx)
                handler.run_migrations(op)
        else:
            with ctx.begin_transaction():
                handler.run_migrations(op)
    except sa.exc.OperationalError as error:
        exception = create_exception(manifest, error)
        raise exception


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, Model, MigrationHandler, MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, old: sa.Table, new: Model,
            handler: MigrationHandler, rename: MigrateRename):
    columns = []
    table_name = rename.get_table_name(old.name)

    # Handle table renames
    for column in old.columns:
        columns.append(rename.get_column_name(old.name, column.name))
    if table_name != old.name:
        handler.add_action(ma.RenameTableMigrationAction(old.name, table_name))
        changelog_name = get_pg_name(f'{old.name}{TableType.CHANGELOG.value}')
        file_name = get_pg_name(f'{old.name}{TableType.FILE.value}')
        if inspector.has_table(changelog_name):
            handler.add_action(ma.RenameTableMigrationAction(
                old_table_name=changelog_name,
                new_table_name=get_pg_name(f'{table_name}{TableType.CHANGELOG.value}')
            ))
        for table in inspector.get_table_names():
            if table.startswith(file_name):
                split = table.split(TableType.FILE.value)
                handler.add_action(ma.RenameTableMigrationAction(
                    old_table_name=file_name,
                    new_table_name=get_pg_name(f'{table_name}{TableType.FILE.value}{split[1]}')
                ))

    # Handle Refs (required because, level < 4 can create multiple columns) and Files
    properties = list(new.properties.values())
    for prop in new.properties.values():
        if isinstance(prop.dtype, (Ref, File)):
            items = []
            column_name = rename.get_old_column_name(old.name, get_column_name(prop))
            for column in old.columns:
                if '.' in column.name and column.name.split('.')[0] == column_name:
                    renamed = rename.get_column_name(old.name, column.name)
                    if renamed in columns:
                        columns.remove(renamed)
                    items.append(column)
            if prop in properties and items:
                properties.remove(prop)
                commands.migrate(context, backend, inspector, old, items, prop.dtype, handler, rename)
        elif isinstance(prop.dtype, Inherit):
            properties.remove(prop)

    props = zipitems(
        columns,
        properties,
        _property_name_key
    )
    for items in props:
        for old_prop, new_prop in items:

            if new_prop and new_prop.name.startswith('_'):
                continue
            if old_prop:
                old_prop = old.columns.get(rename.get_old_column_name(old.name, old_prop))
            if not new_prop:
                commands.migrate(context, backend, inspector, old, old_prop, new_prop, handler, rename, False)
            else:
                commands.migrate(context, backend, inspector, old, old_prop, new_prop, handler, rename)

    required_unique_constraints = []
    if new.unique:
        for val in new.unique:
            prop_list = []

            for prop in val:
                for name in _get_prop_names(prop):
                    prop_list.append(get_pg_name(name))

            constraints = inspector.get_unique_constraints(old.name)
            constraint_name = f'{old.name}_{"_".join(prop_list)}_key'
            if old.name != table_name:
                constraint_name = f'{table_name}_{"_".join(prop_list)}_key'
            constraint_name = get_pg_name(constraint_name)
            required_unique_constraints.append(prop_list)

            # Skip unique properties, since they are already handled in other parts
            if not (len(val) == 1 and val[0].dtype.unique):
                if not any(constraint['name'] == constraint_name for constraint in constraints):
                    handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                        table_name=table_name,
                        constraint_name=constraint_name,
                        columns=prop_list
                    ))

    for prop in new.properties.values():
        if prop.dtype.unique:
            result = commands.prepare(context, backend, prop.dtype)
            if not isinstance(result, list):
                result = [result]
            prop_list = []
            for item in result:
                if isinstance(item, sa.Column):
                    prop_list.append(item.name)
            if prop_list:
                required_unique_constraints.append(prop_list)

    # Clean up old multi constraints for non required models (happens when model.ref is changed)
    # Solo unique constraints are handled with property migrate
    if not old.name.startswith("_"):
        unique_constraints = inspector.get_unique_constraints(old.name)
        for constraint in unique_constraints:
            if constraint["column_names"] not in required_unique_constraints:
                if not handler.has_constraint_been_dropped(constraint["name"]):
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=table_name,
                        constraint_name=constraint["name"]
                    ))


@commands.migrate.register(Context, PostgreSQL, Inspector, NotAvailable, Model, MigrationHandler, MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, old: NotAvailable, new: Model,
            handler: MigrationHandler, rename: MigrateRename):
    table_name = get_pg_name(get_table_name(new))

    pkey_type = commands.get_primary_key_type(context, backend)

    columns = []
    for prop in new.properties.values():
        if not prop.name.startswith("_"):
            if isinstance(prop.dtype, File):
                columns += _handle_new_file_type(context, backend, inspector, prop, pkey_type, handler)
            else:
                cols = commands.prepare(context, backend, prop)
                if isinstance(cols, list):
                    for column in cols:
                        if isinstance(column, sa.Column):
                            columns.append(column)
                else:
                    if isinstance(cols, sa.Column):
                        columns.append(cols)
    handler.add_action(ma.CreateTableMigrationAction(
        table_name=table_name,
        columns=[
            sa.Column('_txn', pkey_type, index=True),
            sa.Column('_created', sa.DateTime),
            sa.Column('_updated', sa.DateTime),
            sa.Column("_id", pkey_type, nullable=False, primary_key=True),
            sa.Column("_revision", sa.Text),
            *columns
        ]

    ))

    if new.unique:
        for val in new.unique:
            prop_list = []
            for prop in val:
                for name in _get_prop_names(prop):
                    prop_list.append(get_pg_name(name))
            constraint_name = get_pg_name(f'{table_name}_{"_".join(prop_list)}_key')
            handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                table_name=table_name,
                constraint_name=constraint_name,
                columns=prop_list
            ))

    _create_changelog_table(context, new, handler, rename)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, NotAvailable, MigrationHandler, MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, old: sa.Table, new: NotAvailable,
            handler: MigrationHandler, rename: MigrateRename):
    old_name = old.name
    remove_name = _get_remove_name(old_name)

    if not old_name.split("/")[-1].startswith("__"):
        if inspector.has_table(remove_name):
            handler.add_action(ma.DropTableMigrationAction(
                table_name=remove_name
            ))
            remove_changelog_name = get_pg_name(f'{remove_name}{TableType.CHANGELOG.value}')
            if inspector.has_table(remove_changelog_name):
                handler.add_action(ma.DropTableMigrationAction(
                    table_name=remove_changelog_name
                ))
            remove_file_name = get_pg_name(f'{remove_name}{TableType.FILE.value}')
            for table in inspector.get_table_names():
                if table.startswith(remove_file_name):
                    handler.add_action(ma.DropTableMigrationAction(
                        table_name=table
                    ))

        handler.add_action(ma.RenameTableMigrationAction(
            old_table_name=old_name,
            new_table_name=remove_name
        ))
        _drop_all_indexes_and_constraints(inspector, old_name, remove_name, handler)

        old_changelog_name = get_pg_name(f'{old_name}{TableType.CHANGELOG.value}')
        new_changelog_name = get_pg_name(f'{remove_name}{TableType.CHANGELOG.value}')
        if inspector.has_table(old_changelog_name):
            handler.add_action(ma.RenameTableMigrationAction(
                old_table_name=old_changelog_name,
                new_table_name=new_changelog_name
            ))
            handler.add_action(ma.RenameSequenceMigrationAction(
                old_name=f'{old_changelog_name}__id_seq',
                new_name=f'{new_changelog_name}__id_seq'
            ))
            _drop_all_indexes_and_constraints(inspector, old_changelog_name, new_changelog_name, handler)
        for table in inspector.get_table_names():
            if table.startswith(f'{old_name}{TableType.FILE.value}'):
                split = table.split(TableType.FILE.value)
                new_name = get_pg_name(f'{remove_name}{TableType.FILE.value}{split[1]}')
                handler.add_action(ma.RenameTableMigrationAction(
                    old_table_name=table,
                    new_table_name=new_name
                ))
                _drop_all_indexes_and_constraints(inspector, table, new_name, handler)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, sa.Column, Property, MigrationHandler,
                           MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: sa.Column, new: Property, handler: MigrationHandler, rename: MigrateRename):
    commands.migrate(context, backend, inspector, table, old, new.dtype, handler, rename)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, NotAvailable, Property, MigrationHandler,
                           MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: NotAvailable, new: Property, handler: MigrationHandler, rename: MigrateRename):
    commands.migrate(context, backend, inspector, table, old, new.dtype, handler, rename)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, NotAvailable, File, MigrationHandler,
                           MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: NotAvailable, new: File, handler: MigrationHandler, rename: MigrateRename):
    name = get_column_name(new.prop)
    nullable = not new.required
    table_name = rename.get_table_name(table.name)
    pkey_type = commands.get_primary_key_type(context, new.backend)
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(f'{name}._id', sa.String, nullable=nullable)
    ))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(f'{name}._content_type', sa.String, nullable=nullable)
    ))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=sa.Column(f'{name}._size', BIGINT, nullable=nullable)
    ))
    if BackendFeatures.FILE_BLOCKS in new.backend.features:
        handler.add_action(ma.AddColumnMigrationAction(
            table_name=table_name,
            column=sa.Column(f'{name}._bsize', sa.Integer, nullable=nullable)
        ))
        handler.add_action(ma.AddColumnMigrationAction(
            table_name=table_name,
            column=sa.Column(f'{name}._blocks', ARRAY(pkey_type, ), nullable=nullable)
        ))
    old_table = get_pg_name(f'{table.name}{TableType.FILE.value}/{name}')
    new_table = get_pg_name(f'{table_name}{TableType.FILE.value}/{name}')
    if not inspector.has_table(old_table):
        handler.add_action(ma.CreateTableMigrationAction(
            table_name=new_table,
            columns=[
                sa.Column('_id', pkey_type, primary_key=True),
                sa.Column('_block', sa.LargeBinary)
            ]
        ))


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, list, File, MigrationHandler, MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: list, new: File, handler: MigrationHandler, rename: MigrateRename):
    column_name = rename.get_column_name(table.name, new.prop.name)
    old_name = rename.get_old_column_name(table.name, new.prop.name)
    for item in old:
        table_name = rename.get_table_name(table.name)
        nullable = new.required if new.required == item.nullable else None
        item_name = item.name
        new_name = item_name.replace(item_name.split(".")[0], column_name) if not item_name.startswith(column_name) else None
        if nullable is not None or new_name is not None:
            handler.add_action(ma.AlterColumnMigrationAction(
                table_name=table_name,
                column_name=item.name,
                nullable=nullable,
                new_column_name=new_name
            ))
    table_name = rename.get_table_name(table.name)
    old_table = get_pg_name(f'{table.name}{TableType.FILE.value}/{old_name}')
    new_table_old_prop = get_pg_name(f'{table_name}{TableType.FILE.value}/{old_name}')
    new_table_new_prop = get_pg_name(f'{table_name}{TableType.FILE.value}/{column_name}')
    if old_name != column_name and inspector.has_table(old_table):
        handler.add_action(
            ma.RenameTableMigrationAction(
                old_table_name=new_table_old_prop,
                new_table_name=new_table_new_prop
            )
        )


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, NotAvailable, Ref, MigrationHandler, MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: NotAvailable, new: Ref, handler: MigrationHandler, rename: MigrateRename):
    columns = commands.prepare(context, backend, new.prop)
    if not isinstance(columns, list):
        columns = [columns]
    for column in columns:
        if isinstance(column, sa.Column):
            commands.migrate(context, backend, inspector, table, old, column, handler, rename, True)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, list, Ref, MigrationHandler, MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: List[sa.Column], new: Ref, handler: MigrationHandler, rename: MigrateRename):
    new_columns = commands.prepare(context, backend, new.prop)
    if not isinstance(new_columns, list):
        new_columns = [new_columns]
    table_name = rename.get_table_name(table.name)
    old_ref_table = rename.get_old_table_name(get_table_name(new.model))
    old_prop_name = rename.get_old_column_name(table.name, get_column_name(new.prop))

    old_names = {}
    new_names = {}
    for item in old:
        base_name = item.name.split(".")
        name = f'{rename.get_column_name(table.name, base_name[0])}.{rename.get_column_name(old_ref_table, base_name[1])}'
        old_names[name] = item
    for item in new_columns:
        new_names[item.name] = item

    if not new.prop.level or new.prop.level > 3:
        if len(old) == 1 and old[0].name == f'{old_prop_name}._id':
            column = old[0]
            new_column = new_columns[0]
            commands.migrate(context, backend, inspector, table, column, new_column, handler, rename, True)
        else:
            column_list = []
            drop_list = []
            for column in old:
                column_list.append(rename.get_column_name(table.name, column.name))
                if column.name != f'{old_prop_name}._id':
                    drop_list.append(column)

            inspector_columns = inspector.get_columns(table.name)
            old_col = NA
            if any(f'{old_prop_name}._id' == col["name"] for col in inspector_columns):
                old_col = table.c.get(f'{old_prop_name}._id')
            commands.migrate(context, backend, inspector, table, old_col, new_columns[0], handler, rename, True)
            if old_names.keys() != new_names.keys():
                handler.add_action(ma.UpgradeTransferDataMigrationAction(
                    table_name=table_name,
                    foreign_table_name=get_table_name(new.refprops[0]),
                    columns=column_list
                ), True)
            for col in drop_list:
                commands.migrate(context, backend, inspector, table, col, NA, handler, rename, True)
    else:
        if len(old) == 1 and old[0].name == f'{old_prop_name}._id':
            requires_drop = True
            for new_column in new_columns:
                if isinstance(new_column, sa.Column):
                    if new_column.name == f'{old_prop_name}._id':
                        requires_drop = False
                        commands.migrate(context, backend, inspector, table, old[0], new_column, handler, rename, True)
                    else:
                        commands.migrate(context, backend, inspector, table, NA, new_column, handler, rename, True)
            names = []
            for item in new_columns:
                names.append(item.name)
            if old_names.keys() != new_names.keys():
                handler.add_action(
                    ma.DowngradeTransferDataMigrationAction(
                        table_name=table_name,
                        foreign_table_name=get_table_name(new.refprops[0]),
                        columns=names
                    ), True
                )
            if requires_drop:
                commands.migrate(context, backend, inspector, table, old[0], NA, handler, rename, True)
        else:
            props = zipitems(
                old_names.keys(),
                new_names.keys(),
                _name_key
            )
            drop_list = []
            for prop in props:
                for old_prop, new_prop in prop:
                    old_prop = old_names[old_prop]
                    new_prop = new_names[new_prop]
                    if old_prop is not None and new_prop is None:
                        drop_list.append(old_prop)
                    else:
                        commands.migrate(context, backend, inspector, table, old_prop, new_prop, handler, rename, True)
            for drop in drop_list:
                commands.migrate(context, backend, inspector, table, drop, NA, handler, rename, True)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, NotAvailable, DataType, MigrationHandler,
                           MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: NotAvailable, new: DataType, handler: MigrationHandler, rename: MigrateRename):
    column = commands.prepare(context, backend, new.prop)
    commands.migrate(context, backend, inspector, table, old, column, handler, rename, False)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, sa.Column, DataType, MigrationHandler,
                           MigrateRename)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: sa.Column, new: DataType, handler: MigrationHandler, rename: MigrateRename):
    column = commands.prepare(context, backend, new.prop)
    commands.migrate(context, backend, inspector, table, old, column, handler, rename, False)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, sa.Column, sa.Column, MigrationHandler,
                           MigrateRename, bool)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: sa.Column, new: sa.Column, handler: MigrationHandler, rename: MigrateRename,
            foreign_key: bool = False):
    column_name = new.name
    table_name = rename.get_table_name(table.name)
    new_type = new.type.compile(dialect=postgresql.dialect())
    old_type = old.type.compile(dialect=postgresql.dialect())
    using = None
    # Convert sa.Float, to postgresql DOUBLE PRECISION type
    if isinstance(new.type, sa.Float):
        new_type = 'DOUBLE PRECISION'
    if isinstance(old.type, geoalchemy2.types.Geometry) and isinstance(new.type, geoalchemy2.types.Geometry):
        if old.type.srid != new.type.srid:
            srid_name = f'"{old.name}"'
            srid = new.type.srid
            if old.type.srid == -1:
                srid_name = f'ST_SetSRID({srid_name}, 4326)'
            if new.type.srid == -1:
                srid = 4326
            using = f'ST_Transform({srid_name}, {srid})'

    nullable = new.nullable if new.nullable != old.nullable else None
    type_ = new.type if old_type != new_type else None
    new_name = column_name if old.name != new.name else None

    if nullable is not None or type_ is not None or new_name is not None or using is not None:
        handler.add_action(ma.AlterColumnMigrationAction(
            table_name=table_name,
            column_name=old.name,
            nullable=nullable,
            type_=type_,
            new_column_name=new_name,
            using=using
        ), foreign_key)

    renamed = _check_if_renamed(table.name, table_name, old.name, new.name)
    unique_name = get_pg_name(f'{table_name}_{column_name}_key')
    removed = []
    if renamed:
        unique_constraints = inspector.get_unique_constraints(table_name=table.name)
        for constraint in unique_constraints:
            if constraint["column_names"] == [old.name]:
                removed.append(constraint["name"])
                handler.add_action(ma.DropConstraintMigrationAction(
                    constraint_name=constraint["name"],
                    table_name=table_name
                ), foreign_key)
                unique_name = get_pg_name(_rename_index_name(constraint["name"], table.name, table_name, old.name, new.name))
                if new.unique:
                    handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                        constraint_name=unique_name,
                        table_name=table_name,
                        columns=[column_name]
                    ), foreign_key)
    else:
        if new.unique:
            if not any(constraint["column_names"] == [column_name] for constraint in inspector.get_unique_constraints(table_name=table.name)) and not any(index["column_names"] == [column_name] and index["unique"] for index in inspector.get_indexes(table_name=table.name)):
                handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                    constraint_name=unique_name,
                    table_name=table_name,
                    columns=[column_name]
                ), foreign_key)
        else:
            for constraint in inspector.get_unique_constraints(table_name=table.name):
                if constraint["column_names"] == [column_name]:
                    # Check if old column was unique, if not add it to removed list, but don't actually remove it, other part of the code handles this case
                    if not old.unique:
                        removed.append(constraint["name"])
                        continue
                    removed.append(constraint["name"])
                    handler.add_action(ma.DropConstraintMigrationAction(
                        constraint_name=constraint["name"],
                        table_name=table_name,
                    ), foreign_key)
            for index in inspector.get_indexes(table_name=table.name):
                if index["column_names"] == [column_name] and index["unique"] and index["name"] not in removed:
                    index_name = index["name"]
                    handler.add_action(ma.DropIndexMigrationAction(
                        index_name=index_name,
                        table_name=table_name,
                    ), foreign_key)
                    handler.add_action(ma.CreateIndexMigrationAction(
                        index_name=index_name,
                        table_name=table_name,
                        columns=[column_name]
                    ), foreign_key)

    index_name = get_pg_name(f'ix_{table_name}_{column_name}')
    if renamed:
        indexes = inspector.get_indexes(table_name=table.name)
        for index in indexes:
            if index["column_names"] == [old.name] and index["name"] not in removed:
                handler.add_action(
                    ma.DropIndexMigrationAction(
                        index_name=index["name"],
                        table_name=table_name
                    ), foreign_key)
                index_name = get_pg_name(_rename_index_name(index["name"], table.name, table_name, old.name, new.name))
                if new.index:
                    handler.add_action(ma.CreateIndexMigrationAction(
                        index_name=index_name,
                        table_name=table_name,
                        columns=[column_name]
                    ), foreign_key)
                elif isinstance(new.type, geoalchemy2.types.Geometry):
                    handler.add_action(ma.CreateIndexMigrationAction(
                        index_name=index_name,
                        table_name=table_name,
                        columns=[column_name],
                        using="gist"
                    ), foreign_key)
    else:
        if new.index:
            if not any(index["column_names"] == [column_name] for index in inspector.get_indexes(table_name=table.name)):
                handler.add_action(ma.CreateIndexMigrationAction(
                    index_name=index_name,
                    table_name=table_name,
                    columns=[column_name]
                ), foreign_key)
        elif isinstance(new.type, geoalchemy2.types.Geometry):
            if not any(index["column_names"] == [column_name] for index in inspector.get_indexes(table_name=table.name)):
                handler.add_action(ma.CreateIndexMigrationAction(
                    index_name=index_name,
                    table_name=table_name,
                    columns=[column_name],
                    using="gist"
                ), foreign_key)
        else:
            for index in inspector.get_indexes(table_name=table.name):
                if index["column_names"] == [column_name] and index["name"] not in removed:
                    if not (index["unique"] and new.unique):
                        handler.add_action(ma.DropIndexMigrationAction(
                            index_name=index["name"],
                            table_name=table_name,
                        ), foreign_key)


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, NotAvailable, sa.Column, MigrationHandler,
                           MigrateRename, bool)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: NotAvailable, new: sa.Column, handler: MigrationHandler, rename: MigrateRename,
            foreign_key: bool = False):
    table_name = get_pg_name(rename.get_table_name(table.name))
    handler.add_action(ma.AddColumnMigrationAction(
        table_name=table_name,
        column=new,
    ), foreign_key)
    if new.unique:
        constraint_name = get_pg_name(f'{table_name}_{new.name}_key')
        if not any(constraint["name"] == constraint_name for constraint in inspector.get_unique_constraints(table_name)):
            handler.add_action(ma.CreateUniqueConstraintMigrationAction(
                constraint_name=constraint_name,
                table_name=table_name,
                columns=[new.name]
            ))
    index_required = isinstance(new.type, geoalchemy2.types.Geometry)
    if index_required:
        handler.add_action(ma.CreateIndexMigrationAction(
            table_name=table_name,
            columns=[new.name],
            index_name=get_pg_name(f'idx_{table_name}_{new.name}'),
            using='gist'
        ))


@commands.migrate.register(Context, PostgreSQL, Inspector, sa.Table, sa.Column, NotAvailable, MigrationHandler,
                           MigrateRename, bool)
def migrate(context: Context, backend: PostgreSQL, inspector: Inspector, table: sa.Table,
            old: sa.Column, new: NotAvailable, handler: MigrationHandler, rename: MigrateRename,
            foreign_key: bool = False):
    if not old.name.startswith("_"):
        table_name = rename.get_table_name(table.name)
        columns = inspector.get_columns(table.name)
        remove_name = get_pg_name(f'__{old.name}')

        if any(remove_name == column["name"] for column in columns):
            handler.add_action(ma.DropColumnMigrationAction(
                table_name=table_name,
                column_name=remove_name
            ), foreign_key)
        handler.add_action(ma.AlterColumnMigrationAction(
            table_name=table_name,
            column_name=old.name,
            new_column_name=remove_name
        ), foreign_key)
        indexes = inspector.get_indexes(table_name=table.name)
        for index in indexes:
            if index["column_names"] == [old.name]:
                handler.add_action(ma.DropIndexMigrationAction(
                    table_name=table_name,
                    index_name=index["name"],
                ), foreign_key)
        if old.unique:
            unique_constraints = inspector.get_unique_constraints(table_name=table.name)
            for constraint in unique_constraints:
                if old.name in constraint["column_names"]:
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=table_name,
                        constraint_name=constraint["name"],
                    ), foreign_key)


def _drop_all_indexes_and_constraints(inspector: Inspector, table: str, new_table: str, handler: MigrationHandler):
    constraints = inspector.get_unique_constraints(table)
    removed = []
    foreign_keys = inspector.get_foreign_keys(table)
    for key in foreign_keys:
        handler.add_action(ma.DropConstraintMigrationAction(
            table_name=new_table,
            constraint_name=key["name"]
        ), True)

    for constraint in constraints:
        removed.append(constraint["name"])
        handler.add_action(
            ma.DropConstraintMigrationAction(
                table_name=new_table,
                constraint_name=constraint["name"]
            )
        )
    indexes = inspector.get_indexes(table)
    for index in indexes:
        if index["name"] not in removed:
            handler.add_action(
                ma.DropIndexMigrationAction(
                    table_name=new_table,
                    index_name=index["name"]
                )
            )


def _clean_up_file_type(inspector: Inspector, models: List[Model], handler: MigrationHandler, rename: MigrateRename):
    allowed_file_tables = []
    existing_tables = []
    for model in models:
        existing_tables.append(rename.get_old_table_name(model.name))
        for prop in model.properties.values():
            if isinstance(prop.dtype, File):
                old_table = rename.get_old_table_name(get_table_name(model))
                old_column = rename.get_old_column_name(old_table, get_column_name(prop))
                allowed_file_tables.append(get_pg_name(f'{old_table}{TableType.FILE.value}/{old_column}'))

    for table in inspector.get_table_names():
        if TableType.FILE.value in table:
            split = table.split(f'{TableType.FILE.value}/')
            if split[0] in existing_tables:
                if table not in allowed_file_tables and not split[1].startswith("__"):
                    new_name = get_pg_name(f'{split[0]}{TableType.FILE.value}/__{split[1]}')
                    if inspector.has_table(new_name):
                        handler.add_action(ma.DropTableMigrationAction(
                            table_name=new_name
                        ))
                    handler.add_action(ma.RenameTableMigrationAction(
                        old_table_name=table,
                        new_table_name=new_name
                    ))
                    _drop_all_indexes_and_constraints(inspector, table, new_name, handler)


def _handle_foreign_key_constraints(inspector: Inspector, models: List[Model], handler: MigrationHandler,
                                    rename: MigrateRename):
    for model in models:
        source_table = get_pg_name(get_table_name(model))
        old_name = get_pg_name(rename.get_old_table_name(source_table))
        foreign_keys = []
        if old_name in inspector.get_table_names():
            foreign_keys = inspector.get_foreign_keys(old_name)
        if model.base:
            referent_table = get_pg_name(get_table_name(model.base.parent))
            if not model.base.level or model.base.level > 3:
                check = False
                fk_name = get_pg_name(f'fk_{referent_table}_id')
                for key in foreign_keys:
                    if key["constrained_columns"] == ["_id"]:
                        if key["name"] == fk_name and key["referred_table"] == referent_table:
                            check = True
                        else:
                            handler.add_action(ma.DropConstraintMigrationAction(
                                table_name=source_table,
                                constraint_name=key["name"]
                            ), True)
                        break
                if not check:
                    handler.add_action(
                        ma.CreateForeignKeyMigrationAction(
                            source_table=source_table,
                            referent_table=referent_table,
                            constraint_name=fk_name,
                            local_cols=["_id"],
                            remote_cols=["_id"]
                        ), True
                    )
            else:
                for key in foreign_keys:
                    if key["constrained_columns"] == ["_id"]:
                        handler.add_action(ma.DropConstraintMigrationAction(
                            table_name=source_table,
                            constraint_name=key["name"]
                        ), True)
                        break
        else:
            for key in foreign_keys:
                if key["constrained_columns"] == ["_id"]:
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=source_table,
                        constraint_name=key["name"]
                    ), True)
                    break

        required_ref_props = {}
        for prop in model.properties.values():
            if isinstance(prop.dtype, Ref):
                if not prop.level or prop.level > 3:
                    name = get_pg_name(f"fk_{source_table}_{prop.name}._id")
                    required_ref_props[name] = {
                        "name": name,
                        "constrained_columns": [f"{prop.name}._id"],
                        "referred_table": get_pg_name(prop.dtype.model.name),
                        "referred_columns": ["_id"]
                    }

        for key in foreign_keys:
            if key["constrained_columns"] != ["_id"]:
                if key["name"] not in required_ref_props.keys():
                    handler.add_action(ma.DropConstraintMigrationAction(
                        table_name=source_table,
                        constraint_name=key["name"]
                    ), True)
                else:
                    constraint = required_ref_props[key["name"]]
                    if key["constrained_columns"] == constraint["constrained_columns"] and key["referred_table"] == \
                        constraint["referred_table"] and key["referred_columns"] == constraint["referred_columns"]:
                        del required_ref_props[key["name"]]
                    else:
                        handler.add_action(ma.DropConstraintMigrationAction(
                            table_name=source_table,
                            constraint_name=key["name"]
                        ), True)

        for prop in required_ref_props.values():
            handler.add_action(ma.CreateForeignKeyMigrationAction(
                source_table=source_table,
                referent_table=prop["referred_table"],
                constraint_name=prop["name"],
                local_cols=prop["constrained_columns"],
                remote_cols=prop["referred_columns"]
            ), True)


def _create_changelog_table(context: Context, new: Model, handler: MigrationHandler, rename: MigrateRename):
    table_name = get_pg_name(get_table_name(new, TableType.CHANGELOG))
    pkey_type = commands.get_primary_key_type(context, new.backend)
    handler.add_action(ma.CreateTableMigrationAction(
        table_name=table_name,
        columns=[
            sa.Column('_id', BIGINT, primary_key=True),
            sa.Column('_revision', sa.String),
            sa.Column('_txn', pkey_type, index=True),
            sa.Column('_rid', pkey_type),
            sa.Column('datetime', sa.DateTime),
            sa.Column('action', sa.String(8)),
            sa.Column('data', JSONB)
        ]
    ))


def _handle_new_file_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any, handler: MigrationHandler) -> list:
    name = get_column_name(prop)
    nullable = not prop.dtype.required
    columns = []
    columns += [
        sa.Column(f'{name}._id', sa.String, nullable=nullable),
        sa.Column(f'{name}._content_type', sa.String, nullable=nullable),
        sa.Column(f'{name}._size', BIGINT, nullable=nullable)
    ]
    if BackendFeatures.FILE_BLOCKS in prop.dtype.backend.features:
        columns += [
            sa.Column(f'{name}._bsize', sa.Integer, nullable=nullable),
            sa.Column(f'{name}._blocks', ARRAY(pkey_type, ), nullable=nullable),
        ]
    new_table = get_pg_name(get_table_name(prop, TableType.FILE))
    if not inspector.has_table(new_table):
        handler.add_action(ma.CreateTableMigrationAction(
            table_name=new_table,
            columns=[
                sa.Column('_id', pkey_type, primary_key=True),
                sa.Column('_block', sa.LargeBinary)
            ]
        ))
    return columns


def _handle_new_array_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any, handler: MigrationHandler):
    columns = []
    if isinstance(prop.dtype, Array) and prop.dtype.items:
        if prop.list is None:
            columns.append(sa.Column(prop.place, JSONB))

        if isinstance(prop.dtype.items.dtype, File):
            new_columns = _handle_new_file_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        elif isinstance(prop.dtype.items.dtype, Array):
            new_columns = _handle_new_array_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        elif isinstance(prop.dtype.items.dtype, Object):
            new_columns = _handle_new_object_type(context, backend, inspector, prop.dtype.items, pkey_type, handler)
        else:
            new_columns = commands.prepare(context, backend, prop.dtype.items)
        if not isinstance(new_columns, list):
            new_columns = [new_columns]
        for column in new_columns:
            if not isinstance(column, sa.Column):
                new_columns.remove(column)

        new_table = get_pg_name(get_table_name(prop, TableType.LIST))
        if not inspector.has_table(new_table):
            main_table_name = get_pg_name(get_table_name(prop.model))
            handler.add_action(ma.CreateTableMigrationAction(
                table_name=new_table,
                columns=[
                    sa.Column('_txn', pkey_type, index=True),
                    sa.Column('_rid', pkey_type, sa.ForeignKey(
                        f'{main_table_name}._id', ondelete='CASCADE',
                    ), index=True),
                    *new_columns
                ]
            ))
    return columns


def _handle_new_object_type(context: Context, backend: PostgreSQL, inspector: Inspector, prop: Property, pkey_type: Any, handler: MigrationHandler):
    columns = []
    if isinstance(prop.dtype, Object) and prop.dtype.properties:
        for new_prop in prop.dtype.properties.values():
            if prop.name.startswith('_') and prop.name not in ('_revision',):
                continue
            if isinstance(new_prop.dtype, File):
                columns = _handle_new_file_type(context, backend, inspector, new_prop, pkey_type, handler)
            elif isinstance(new_prop.dtype, Array):
                columns = _handle_new_array_type(context, backend, inspector, new_prop, pkey_type, handler)
            elif isinstance(new_prop.dtype, Object):
                columns = _handle_new_object_type(context, backend, inspector, new_prop, pkey_type, handler)
            else:
                columns = commands.prepare(context, backend, new_prop)

            if not isinstance(columns, list):
                columns = [columns]
            for column in columns:
                if not isinstance(column, sa.Column):
                    columns.remove(column)
    return columns


def _check_if_renamed(old_table: str, new_table: str, old_property: str, new_property:str):
    return old_table != new_table or old_property != new_property


def _rename_index_name(index: str, old_table: str, new_table: str, old_property: str, new_property:str):
    new = index.replace(old_table, new_table)
    new = new.replace(old_property, new_property)
    return new


def _model_name_key(model: str) -> str:
    return get_pg_name(model)


def _property_name_key(prop: Any) -> str:
    name = prop
    if isinstance(prop, Property):
        name = get_column_name(prop)
    return get_pg_name(name).split(".")[0]


def _name_key(name: str):
    return name


def _get_remove_name(name: str) -> str:
    new_name = name.split("/")
    if not new_name[-1].startswith("__"):
        new_name[-1] = f'__{new_name[-1]}'
    new_name = '/'.join(new_name)
    new_name = get_pg_name(new_name)
    return new_name


def _get_prop_names(prop: Property):
    name = prop.name
    if isinstance(prop.dtype, Ref):
        if not prop.level or prop.level > Level.open:
            name = f'{name}._id'
        else:
            for refprop in prop.dtype.refprops:
                yield f'{name}.{refprop.name}'
    yield name



import ujson as json

from spinta.formats.components import Format


class JsonLines(Format):
    content_type = 'application/x-json-stream'
    accept_types = {
        'application/x-json-stream',
    }
    params = {}

    def __call__(self, data):
        memory = next(data)
        for row in data:
            yield json.dumps({k: v for k, v in memory.items() if k != '_page'}, ensure_ascii=False) + '\n'
            memory = row

        last_row = {}
        for k, v in memory.items():
            if k == '_page':
                last_row[k] = {
                    'next': v
                }
            else:
                last_row[k] = v
        yield json.dumps(last_row, ensure_ascii=False) + '\n'
